/*  voodooacpi.go - ACPI Control daemon written in Go
    Copyright (C) 2023  hoff.industries

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>. */

package main

import (
	"os"
	"os/exec"
	"fmt"
	"hoff.industries/marty/libmmp"
	"hoff.industries/marty/libmarty"
	"hoff.industries/marty/mendpoints"
)

var TARGET_PROTOCOL = []int{2, 0, 0}
var INSTANCE_VERSION = []int{3, 0, 0}

const INSTANCE_NAME string = "m-voodooacpi"
const INSTANCE_TITLE string = "VoodooACPI Control Daemon"

var Endpoints = []libmarty.MartyClientEndpoint{

	{
		Endpoint: mendpoints.VOODOO_EPT_SLEEP,
		Name: "sleep",
		Min: 5,
		Max: 5,
		IsAscii: false,
	},

	{
		Endpoint: mendpoints.VOODOO_EPT_SHUTDOWN,
		Name: "shutdown",
		Min: 5,
		Max: 5,
		IsAscii: false,
	},

	{
		Endpoint: mendpoints.VOODOO_EPT_WAKE,
		Name: "wake",
		Min: 6,
		Max: 6,
		IsAscii: true,
	},
}

var MaxEndpoint = len(Endpoints)

var MAC_CONFIG map[string]string = map[string]string{
	"sunyshore": "00:1f:bc:13:54:30",
	"blackthorn": "08:be:ac:22:8c:3c",
}

func main() {
	// define standard base structures for my marty clientship
	var data libmarty.MartyClientData
	var status libmmp.MMStatus

	var txPacket libmmp.MMPacket
	var rxPacket *libmmp.MMPacket = nil

	var applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["SUCCESS"]

	var hostname string = ""
	var callError error = nil

	var endpointConversion int = -1
	var endpoint *libmarty.MartyClientEndpoint = nil

	var mapLookup bool = false

	var macAddress string
	var execCommand string
	var execArgs []string

	var execution *exec.Cmd = nil

	libmarty.Marty_client_data_factory(&data)

	libmmp.Mmstatus_factory(&status)
	libmmp.Mmpacket_factory(&txPacket, true)

	// set up initial client settings
	data.VersionString = fmt.Sprintf("%s v%d.%d.%d", INSTANCE_NAME,
		INSTANCE_VERSION[0], INSTANCE_VERSION[1], INSTANCE_VERSION[2])

	fmt.Printf("%s[%s]%s Starting %s %s - %s\n",
		libmarty.LoggerColours[libmarty.LOG_INFO], libmarty.MartyFlavour,
			libmarty.ClearColour, libmarty.LIBMARTY_AUTHOR, INSTANCE_TITLE, data.VersionString)

	hostname, callError = os.Hostname()
	if (callError != nil) {

		os.Exit(libmarty.EXIT_ERR_HOSTNAME)
	}

	data.ConnType = libmarty.CONN_UNIX
	data.InitParameters.Auth = ""
	data.InitParameters.Priority = 0
	data.InitParameters.Host = hostname
	data.InitParameters.Name = "voodooacpi"

	if (libmarty.Spawn_marty_socket(&data) != libmmp.NONE) {

		libmarty.Marty_full_log("Socket spawn failure!\n",
			libmarty.LOG_CRITICAL, &data)
		os.Exit(libmarty.EXIT_ERR_SPAWN)
	}

	if (libmarty.Marty_init(&data) != libmmp.NONE) {

		libmarty.Marty_full_log("MMP initialisation failure!\n",
			libmarty.LOG_CRITICAL, &data)
		os.Exit(libmarty.EXIT_ERR_INIT)
	};

	libmarty.Marty_full_log("MMP initialisation success!\n", libmarty.LOG_STARTUP, &data)

	go libmarty.Marty_network_loop(&data)
	libmarty.Marty_full_log("Network loop started.\n", libmarty.LOG_STARTUP, &data)

	for true {

		libmmp.Mmpacket_reset(&txPacket, true)
		applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["SUCCESS"]

		if (rxPacket != nil) {
			libmmp.Mmpacket_reset(rxPacket, true)
		}

		rxPacket = nil

		select {
			//case deliveryHandle = <- data.DeliveryChannel:
			case _ = <- data.DeliveryChannel:
				continue
				// listen for new SEND messages from client loop on channel
			case rxPacket = <- data.MessageChannel:
		}

		if (rxPacket == nil) {
			libmarty.Marty_full_log("Received NULL pointer from ITC channel!\n", libmarty.LOG_WARNING, &data)
			continue
		}

		endpointConversion = int(rxPacket.Payloads[libmarty.PAYL_IDX_ENDPOINT][0])

		if (endpointConversion > MaxEndpoint || endpointConversion < 0) {

			libmarty.Marty_full_log("Requested endpoint outside my range!\n", libmarty.LOG_WARNING, &data)
			applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["ENDPOINT"]

			goto JMP_TX_STATUS
		}

		endpoint = &Endpoints[endpointConversion]

		if (len(rxPacket.Payloads) < endpoint.Min) {
			libmarty.Marty_full_log("Request payload count less than expected!\n", libmarty.LOG_WARNING, &data)

			applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["TOO_FEW"]

			goto JMP_TX_STATUS
		}

		if (len(rxPacket.Payloads) > endpoint.Max) {
			libmarty.Marty_full_log("Request payload count more than expected!\n", libmarty.LOG_WARNING, &data)

			applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["TOO_MANY"]

			goto JMP_TX_STATUS
		}

		switch (endpointConversion) {
			case mendpoints.VOODOO_EPT_SLEEP:
				execCommand = "loginctl"
				execArgs = []string{"suspend"}

			case mendpoints.VOODOO_EPT_SHUTDOWN:
				execCommand = "poweroff"
				execArgs = []string{}

			case mendpoints.VOODOO_EPT_WAKE:

				macAddress, mapLookup = MAC_CONFIG[libmmp.Payload_2_str(rxPacket.Payloads[5])]

				if (!mapLookup) {
					continue
				}

				execCommand = "wakeonlan"
				execArgs = []string{"-i", "192.168.1.255", macAddress}
		}

		execution = exec.Command(execCommand, execArgs...)
		execution.Start()

		JMP_TX_STATUS:

		libmarty.Marty_set_application(&txPacket, &status, rxPacket.Payloads[0], applicationStatus)
		if (status.Error != libmmp.NONE) {
			libmarty.Marty_full_log("Couldn't set autopacket!\n", libmarty.LOG_ERROR, &data)
			continue
		}

		libmmp.Serialize(&txPacket, &status)
		if (status.Error != libmmp.NONE) {
			libmarty.Marty_full_log("Couldn't serialize autopacket!\n", libmarty.LOG_ERROR, &data)
			fmt.Printf("%d\n", status.Error)
			continue
		}

		libmarty.Marty_net_write(&txPacket, &status, &data)
	}
}
