""" libmmp.py - MARTY Middleware Protocol Library written in Python
    Copyright (C) 2023  Matthew Hoffman

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>."""

from enum import Enum
from sys import byteorder

MMP_VERSION = "0.0.1"
IMPLEMENTS = "2.0"

SONAR_RETRIES = 3
SONAR_SLEEP_INTERVAL = 5

MAX_SIZE = 4096
MAX_CLIENTS = 20

DATA_EMPTY = b'\x00'
ARGS_EMPTY = b'\x00\x00'
TERM_PAYLOAD = b'\xFE'
TERM_PACKET = b'\xFF'

PKT_IDX_PAYLOAD_START = 3

PAYLOAD_FIELD_OPT = 0
PAYLOAD_FIELD_OPT_R = 1
PAYLOAD_FIELD_REQ = 2
PAYLOAD_FIELD_NULL = 3

class MMPError(Enum):
	NONE = 0
	NETWORK = 1
	INIT_FAIL = 2
	DELIVERY_FAIL = 3
	QUERY_FAIL = 4
	QUEUE_PUSH = 5
	QUEUE_PULL = 6
	UNSET_MESSAGE_TYPE = 7
	UNSET_ARGUMENT = 8
	UNSET_PAYLOADS = 9
	UNSET_PACKET = 10
	INVALID_TYPE = 11
	INVALID_PACKET_LENGTH = 12
	INVALID_PAYLOAD_COUNT = 13
	INVALID_PAYLOAD_VALUE = 14
	INVALID_MAP_KEY = 15
	MESSAGE_ID_UNAVAILABLE = 16
	WAIT_INTERRUPTION = 17
	AUTH_FAIL = 18

class ClientState(Enum):
	PREINIT = 0
	ACTIVE = 1
	INACTIVE = 2

class PayloadConfiguration(Enum):
	OPT = 0
	OPT_R = 1
	REQ = 2

class MessageDirection(Enum):
	FROM_CLIENT = 0
	FROM_SERVER = 1

MESSAGE_TYPES = {
	"NULL": b'\x00',
	"INIT": b'\x01',
	"INVALID": b'\x02',
	"CONTERM": b'\x03',
	"SONAR": b'\x04',
	"SEND": b'\x05',
	"APPLICATION": b'\x06',
	"QUERY": b'\x07',
}

PAYLOAD_COUNT = {
	"NULL": (0, False),
	"INIT": (4, False),
	"INVALID": (1, False),
	"CONTERM": (0, False),
	"SONAR": (0, False),
	"SEND": (5, True),
	"APPLICATION": (1, False),
	"QUERY": (3, False),
}

BYTE_TO_TYPE = {
	b'\x00': "NULL",
	b'\x01': "INIT",
	b'\x02': "INVALID",
	b'\x03': "CONTERM",
	b'\x04': "SONAR",
	b'\x05': "SEND",
	b'\x06': "APPLICATION",
	b'\x07': "QUERY",
}

ARGS_FROM_CLIENT = {
	"INIT": {
		"INIT": b'\x00\x01',
	},
}

ARGS_FROM_SERVER = {
	"INIT": {
		"ACCEPT": b'\x00\x01',
	},
	"INVALID": {
		"AUTH": b'\x00\x06',
		"KEY_EXISTS": b'\x00\x07',
		"RESERVED": b'\x00\x08',
		"MESSAGE_ID": b'\x00\x09',
	},
	"CONTERM": {
		"AUTH": b'\x00\x03',
	},
	"QUERY": {
		"NOT_FOUND": b'\x00\x03',
	},
}

ARGS_BIDIRECTIONAL = {
	"NULL": {
		"EMPTY": b'\x00\x00',
	},
	"INVALID": {
		"TYPE": b'\x00\x01',
		"LENGTH": b'\x00\x02',
		"ARGS": b'\x00\x03',
		"PAYLOAD": b'\x00\x04',
		"PAYLOAD_COUNT": b'\x00\x05',
	},
	"CONTERM": {
		"CLEAN": b'\x00\x01',
		"SPAM": b'\x00\x02',
	},
	"SONAR": {
		"EMPTY": b'\x00\x00',
	},
	"SEND": {
		"DIRECT": b'\x00\x01',
		"BROADCAST": b'\x00\x02',
	},
	"APPLICATION": {
		"SUCCESS": b'\x00\x01',
		"DELIVERY": b'\x00\x02',
		"BUSY": b'\x00\x03',
		"RANGE": b'\x00\x04',
		"ENDPOINT": b'\x00\x05',
		"VALUE": b'\x00\x06',
		"TOO_FEW": b'\x00\x07',
		"TOO_MANY": b'\x00\x08',
		"AUTH": b'\x00\x09',
		"EXECUTE": b'\x00\x0A',
	},
	"QUERY": {
		"CONN_NUMBER": b'\x00\x01',
		"CLIENT_INFO": b'\x00\x02',
	},
}

class MMPacket():
	def __init__(self, message_type="NULL",
			argument=ARGS_EMPTY, payloads=None, direction=MessageDirection.FROM_CLIENT):

		self._message_type = None

		if isinstance(message_type, str):
			try:
				self._message_type = MESSAGE_TYPES[message_type]
			except KeyError:
				self._message_type = MESSAGE_TYPES["NULL"]

		elif isinstance(message_type, bytes):
			self._message_type = message_type

		self._message_type_str = None

		if self._message_type is not None:
			self._message_type_str = BYTE_TO_TYPE[self._message_type]

		self._argument = None

		if isinstance(argument, str):
			if direction == MessageDirection.FROM_CLIENT:
				try:
					self._argument = ARGS_FROM_CLIENT[self._message_type_str][argument]
				except KeyError:
					pass

			elif direction == MessageDirection.FROM_SERVER:
				try:
					self._argument = ARGS_FROM_SERVER[self._message_type_str][argument]
				except KeyError:
					pass

			else:
				pass

			if self._argument is None:
				try:
					self._argument = ARGS_BIDIRECTIONAL[self._message_type_str][argument]
				except KeyError:
					pass

		elif isinstance(argument, bytes):
			self._argument = argument

		if self._argument is None:
			self._argument = ARGS_BIDIRECTIONAL["NULL"]["EMPTY"]

		self._payloads = None

		if isinstance(payloads, list):
			self._payloads = payloads.copy()

		self._bytez = None

	def get_message_type(self) -> str:
		return self._message_type

	def get_argument(self) -> bytes:
		return self._argument

	def get_payloads(self) -> bytearray:
		return self._payloads

	def get_bytes(self) -> bytes:
		return self._bytez

	def set_message_type(self, message_type: str):
		self._message_type = message_type

	def set_argument(self, argument: bytes):
		self._argument = argument

	def set_payloads(self, payloads: bytearray):
		self._payloads = payloads

	def construct_packet(self) -> MMPError:
		try:
			payload_count = PAYLOAD_COUNT[self._message_type_str]
		except KeyError:
			return MMPError.INVALID_MAP_KEY

		if self._payloads is None and payload_count[0] != 0:
			return MMPError.UNSET_PAYLOADS

		if self._payloads is not None and len(self._payloads) != payload_count[0]:
			if not payload_count[1] or (payload_count[1] and (len(self._payloads) < payload_count[0])):
				return MMPError.INVALID_PAYLOAD_COUNT

		return_packet = bytearray()

		return_packet.extend(self._message_type)
		return_packet.extend(self._argument)

		if self._payloads is not None:
			for payload in self._payloads:
				return_packet.extend(payload)
				return_packet.extend(TERM_PAYLOAD)

		return_packet.extend(TERM_PACKET)

		self._bytez = return_packet

		return MMPError.NONE

def packet_2_heap(packet: MMPacket) -> MMPacket:
	heap_packet = MMPacket(packet.get_message_type(), packet.get_argument())

	if isinstance(packet.get_payloads(), list):
		heap_packet.set_payloads(packet.get_payloads().copy())

	return heap_packet

def bytelist(input_bytes: bytearray) -> list:
    """ convert input bytearray to python list of bytes """
    return [int.to_bytes(byte, length=1, byteorder=byteorder) for byte in list(input_bytes)]

def extract_arguments(received_bytelist) -> bytes:
    """ extract argument bytes from byte list """
    if len(received_bytelist) < 3:
        return None
    return b''.join([received_bytelist[1], received_bytelist[2]])

def extract_payloads(received_bytelist: list) -> list:
    """ extract payloads from received byte list """
    # return if bad type
    if not isinstance(received_bytelist, list):
        return None

    payloads = []

    # message has no payloads, return empty
    if len(received_bytelist) < 6:
        return []

    payload = bytearray()
    # loop through payload bytes
    for byte in received_bytelist[3:]:
        # if there's a wrong type in input, return none
        if not isinstance(byte, bytes):
            return None

        # stop reading if end of packet
        if byte == TERM_PACKET:
            break
        # append payload if terminated
        if byte == TERM_PAYLOAD:
            payloads.append(payload)
            # reset running payload
            payload = bytearray()
            continue
        # add byte to running payload
        payload.extend(byte)

    return payloads
